@@ -4,12 +4,14 @@
 pub use congestion_controller::CongestionController;
 pub use cubic::CubicCongestionController;
 pub use rtt_estimator::*;
+pub use sent_packets::*;
 
 pub mod congestion_controller;
 pub mod cubic;
 mod hybrid_slow_start;
 mod pacing;
 mod rtt_estimator;
+mod sent_packets;
 
 //= https://www.rfc-editor.org/rfc/rfc9002#section-7.7
 //# Senders SHOULD limit bursts to the initial congestion window; see
